import redis
import json

class KnowledgeManager:
    def __init__(self, config):
        """
        Initialize the Knowledge object.
        :param config: Configuration for knowledge storage (local or global)
        """
        self.config = config
        self.storage_type = config.get('storage_type', 'local')  # 'local' or 'global'

        if self.storage_type == 'global':
            self.redis_client = redis.StrictRedis(
                host=config['redis_host'], 
                port=config['redis_port'], 
                db=config.get('redis_db', 0)
            )
        else:
            self.local_store = {}

    def write(self, key, value):
        """
        Store a key-value pair in the knowledge base. If the value is a dictionary, it will be serialized.
        :param key: The key for storing the data
        :param value: The value to store (can be a dict, str, int, etc.)
        """
        if isinstance(value, dict):
            value = json.dumps(value)  # Serialize the dictionary to a JSON string
        if self.storage_type == 'global':
            self.redis_client.set(key, value)
        else:
            self.local_store[key] = value

    def read(self, key, queueSize = 1):
        """
        Retrieve a value from the knowledge base. If the value is JSON, it will be deserialized to a Python object.
        :param key: The key for retrieving the data
        :return: The value associated with the key or None if not found
        """
        if self.storage_type == 'global':
            value = self.redis_client.get(key)
            if value is not None:
                value = value.decode('utf-8')  # Convert bytes to string
                try:
                    value = json.loads(value)  # Try to deserialize the value if it's a JSON string
                except json.JSONDecodeError:
                    pass  # If it's not JSON, return it as a string
            return value
        else:
            return self.local_store.get(key, None)

    def delete(self, key):
        """
        Delete a key-value pair from the knowledge base.
        :param key: The key to delete
        """
        if self.storage_type == 'global':
            self.redis_client.delete(key)
        else:
            if key in self.local_store:
                del self.local_store[key]

    def exists(self, key):
        """
        Check if a key exists in the knowledge base.
        :param key: The key to check
        :return: True if the key exists, False otherwise
        """
        if self.storage_type == 'global':
            return self.redis_client.exists(key)
        else:
            return key in self.local_store



#**********************************************************************************
# * Copyright (C) 2024-present Bert Van Acker (B.MKR) <bert.vanacker@uantwerpen.be>
# *
# * This file is part of the roboarch R&D project.
# *
# * RAP R&D concepts can not be copied and/or distributed without the express
# * permission of Bert Van Acker
# **********************************************************************************
import json

class namedObject(object):

    def __init__(self, name='tbd', description='tbd', verbose=False):
        self._name = name
        self._description = description
        self._verbose = verbose

    @property
    def name(self):
        """The name property (read-only)."""
        return self._name

    @property
    def description(self):
        """The description property (read-only)."""
        return self._description




class process(namedObject):

    def __init__(self, name='tbd',description='tbd',verbose=False,featureList = None,threadList=None,formalism="python",containerization=False):
        super().__init__(name=name, description=description, verbose=verbose)

        if featureList is not None:
            self._featureList = featureList
        else:
            self._featureList = []

        if threadList is not None:
            self._threadList = threadList
        else:
            self._threadList = []

        self._formalism = formalism
        self._containerization = containerization

    @property
    def features(self):
        return self._featureList

    def addFeature(self, feature):
        """Add a feature to the system """
        self._featureList.append(feature)

    @property
    def threads(self):
        return self._threadList

    def addThread(self, t):
        """Add a thread to the system """
        self._threadList.append(t)

    @property
    def formalism(self):
        return self._formalism

    @formalism.setter
    def formalism(self,f):
        self._formalism = f

    @property
    def containerization(self):
        return self._containerization

    @containerization.setter
    def containerization(self,c):
        self._containerization = c

class thread(namedObject):

    def __init__(self, name='tbd',description='tbd',verbose=False,featureList = None,eventTrigger=None):
        super().__init__(name=name, description=description, verbose=verbose)

        if featureList is not None:
            self._featureList = featureList
        else:
            self._featureList = []

        self._eventTrigger = eventTrigger

    @property
    def features(self):
        return self._featureList
    def addFeature(self, feature):
        """Add a feature to the system """
        self._featureList.append(feature)

    @property
    def eventTrigger(self):
        return self._eventTrigger

class characteristic(namedObject):

    def __init__(self, name='tbd',description='tbd',verbose=False,value="",dataType='-'):
        super().__init__(name=name, description=description, verbose=verbose)

        self._value = value
        self._dataType = dataType

class event(namedObject):

    def __init__(self, name='tbd',description='tbd',verbose=False):
        super().__init__(name=name, description=description, verbose=verbose)

class mode(namedObject):

    def __init__(self, name='tbd',description='tbd',verbose=False,callback = None):
        super().__init__(name=name, description=description, verbose=verbose)

        self._callback = callback

    @property
    def callback(self):
        """The callback property (read-only)."""
        return self._callback

class transition(namedObject):

    def __init__(self, name='tbd',description='tbd',verbose=False,callback = None,source=None,destination=None):
        super().__init__(name=name, description=description, verbose=verbose)

        self._source = source
        self._destination = destination
        self._callback = callback

    @property
    def source(self):
        """The source (read-only)."""
        return self._source

    @property
    def destination(self):
        """The source (read-only)."""
        return self._destination

    @property
    def callback(self):
        """The callback property (read-only)."""
        return self._callback

class feature(namedObject):

    def __init__(self, name='tbd',description='tbd',verbose=False,featureType="feature"):
        super().__init__(name=name, description=description, verbose=verbose)

        self._featureType = featureType

    @property
    def featureType(self):
        return self._featureType

    @featureType.setter
    def featureType(self,type):
        self._featureType = type

class port(feature):

    def __init__(self, name='tbd',description='tbd',initialValue = 1.0,valueReference=1,type='data',message=None,verbose=False):
        super().__init__(name=name, description=description, verbose=verbose)

        self._featureType = 'port'
        self._initialValue = initialValue
        self._valueReference = valueReference
        self._type = type
        self._message = message


    @property
    def value(self):
        """The value property (read)."""
        return self._initialValue

    @value.setter
    def value(self,value):
        """The value property (write)."""
        self._initialValue = value

    @property
    def valueReference(self):
        """The valueReference property (read)."""
        return self._valueReference

    @valueReference.setter
    def valueReference(self, value):
        """The valueReference property (write)."""
        self._valueReference = value

    @property
    def type(self):
        """The type property (read-only)."""
        return self._type

    @property
    def message(self):
        """The message property"""
        return self._message

class inport(port):
    def __init__(self, name='tbd',description='tbd',initialValue = 1.0,valueReference=1,type='data',verbose=False,message=None):
        super().__init__(name=name, description=description, verbose=verbose,initialValue=initialValue,valueReference=valueReference,type=type,message=message)

        self._featureType = 'inport'
class outport(port):
    def __init__(self, name='tbd',description='tbd',initialValue = 1.0,valueReference=1,type='data',verbose=False,message=None):
        super().__init__(name=name, description=description, verbose=verbose,initialValue=initialValue,valueReference=valueReference,type=type,message=message)

        self._featureType = 'outport'

class connection(namedObject):

    def __init__(self, name='tbd',description='tbd',verbose=False,source=None,destination=None):
        super().__init__(name=name, description=description, verbose=verbose)

        self._source = source
        self._destination = destination

    @property
    def source(self):
        """The source (read-only)."""
        return self._source

    @property
    def destination(self):
        """The source (read-only)."""
        return self._destination

class message(namedObject):

    def __init__(self, name='tbd',description='tbd',verbose=False,featureList = None):
        super().__init__(name=name, description=description, verbose=verbose)

        if featureList is not None:
            self._featureList = featureList
        else:
            self._featureList = []

    @property
    def features(self):
        return self._featureList

    def addFeature(self, feature):
        """Add a feature to the system """
        self._featureList.append(feature)

class data(feature):

    def __init__(self, name='tbd',description='tbd',dataType='Float_64',verbose=False):
        super().__init__(name=name, description=description, verbose=verbose)

        self._dataType = dataType

    @property
    def dataType(self):
        return self._dataType

class processor(namedObject):

    def __init__(self,name='tbd', description='tbd', verbose=False,featureList = None,propertyList = None, bindingList=None,IP="localhost"):
        super().__init__(name=name,description = description,verbose=verbose)

        if featureList is not None:
            self._featureList = featureList
        else:
            self._featureList = []
        if propertyList is not None:
            self._propertyList = propertyList
        else:
            self._propertyList = []
        if bindingList is not None:
            self._bindingList = bindingList
        else:
            self._bindingList = []
        self._IP = IP
        self.rap_backbone = False


    def addFeature(self, feature):
        """Add a feature to the system """
        self._featureList.append(feature)

    def addProperty(self, p):
        """Add a property to the system """
        self._propertyList.append(p)

    def addProcessorBinding(self,process):
        """Add a processor binding to the processor """
        self._bindingList.append(process)

    @property
    def processorBinding(self):
        return self._bindingList

    @property
    def IP(self):
        return self._IP

    @IP.setter
    def IP(self,ip):
        self._IP = ip

    @property
    def runs_rap_backbone(self):
        return self.rap_backbone

    @runs_rap_backbone.setter
    def runs_rap_backbone(self, flag):
        self.rap_backbone = flag



class memory(namedObject):

    def __init__(self,name='tbd', description='tbd', verbose=False,propertyList = None):
        super().__init__(name=name,description = description,verbose=verbose)

        if propertyList is not None:
            self._propertyList = propertyList
        else:
            self._propertyList = []

    def addProperty(self, p):
        """Add a property to the system """
        self._propertyList.append(p)

class bus(namedObject):

    def __init__(self,name='tbd', description='tbd', verbose=False,propertyList = None):
        super().__init__(name=name,description = description,verbose=verbose)

        if propertyList is not None:
            self._propertyList = propertyList
        else:
            self._propertyList = []

        self._connectionList = []

    def addProperty(self, p):
        """Add a property to the bus """
        self._propertyList.append(p)

    def addConnection(self, p):
        """Add a connection to the bus """
        self._connectionList.append(p)


class system(namedObject):

    def __init__(self,name='tbd', description='tbd', verbose=False,systemList = None,processList = None,featureList=None,messageList=None,processorList=None,JSONDescriptor=None):
        super().__init__(name=name,description = description,verbose=verbose)

        if featureList is not None:
            self._featureList = featureList
        else:
            self._featureList = []
        if systemList is not None:
            self._systemList = systemList
        else:
            self._systemList = []
        if processList is not None:
            self._processList = processList
        else:
            self._processList = []
        if messageList is not None:
            self._messageList = messageList
        else:
            self._messageList = []
        if processorList is not None:
            self._processorList = processorList
        else:
            self._processorList = []


        if JSONDescriptor is not None:
            self.json2object(JSONDescriptor=JSONDescriptor)

    def addProcess(self,process):
        """Add a process to the process list """
        self._processList.append(process)

    @property
    def processes(self):
        return self._processList

    @property
    def systems(self):
        return self._systemList
    def addSystem(self,system):
        """Add a system to the system list """
        self._systemList.append(system)

    def addFeature(self,feature):
        """Add a feature to the system """
        self._featureList.append(feature)

    def addMessage(self,message):
        """Add a process to the process list """
        self._messageList.append(message)

    @property
    def messages(self):
        return self._messageList
    @messages.setter
    def messages(self,d):
        self._messageList = d

    def addProcessor(self,processor):
        """Add a processor to the system list """
        self._processorList.append(processor)

    @property
    def processors(self):
        return self._processorList

    def object2json(self,fileName):
        """
               Function to generate a json file
        """
        data = json.dumps(self, default=lambda o: o.__dict__, indent=4)
        with open(fileName, 'w', encoding='utf-8') as f:
            f.write(data)

    def json2object(self, JSONDescriptor='system.json'):
        """
             Function to generate an AADLIL system from a JSON file

             :param string JSONDescriptor: absolute path to the json file for the AADLIL system

        """
        # --interpret JSON file--
        with open(JSONDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)
        # --setup object--
        self._name = jsonObject['_name']
        self._description = jsonObject['_description']
        # -- load messages --
        for m in jsonObject['_messageList']:
            features = []
            for d in m['_featureList']:
                tempD = data(name=d['_name'], dataType=d['_dataType'])
                features.append(tempD)
            tempMessage = message(name=m['_name'], featureList=features)
            self.addMessage(tempMessage)
        # -- load systems and containing processes --
        for s in jsonObject['_systemList']:
            tempS=system(name=s['_name'], description=s['_description'])
            for p in s['_processList']:
                tempP=process(name=p['_name'], description=p['_description'])
                features = []
                for f in p['_featureList']:
                    _m = None
                    for m in self.messages:
                        if f['_message'] is not None:
                            if m.name == f['_message']['_name']:
                                _m = m
                    if f['_featureType'] == 'inport':
                        tempF = inport(name=f['_name'], type=f['_type'], message=_m)
                    elif f['_featureType'] == 'outport':
                        tempF = outport(name=f['_name'], type=f['_type'], message=_m)
                    else:
                        tempF = ''
                    tempP.addFeature(tempF)
                threads = []
                for t in p['_threadList']:
                    Tfeatures = []
                    for f in t['_featureList']:
                        _m = None
                        for m in self.messages:
                            if f['_message'] is not None:
                                if m.name == f['_message']['_name']:
                                    _m = m
                        if f['_featureType'] == 'inport':
                            tempF = inport(name=f['_name'], type=f['_type'], message=_m)
                        elif f['_featureType'] == 'outport':
                            tempF = outport(name=f['_name'], type=f['_type'], message=_m)
                        else:
                            tempF = ''
                        Tfeatures.append(tempF)
                    tempT = thread(name=t["_name"],featureList=Tfeatures,eventTrigger=t["_eventTrigger"])
                    tempP.addThread(tempT)
                tempP.formalism=p['_formalism']
                tempP.containerization=p['_containerization']
                tempS.addProcess(tempP)
            self.addSystem(tempS)
        # -- load processors --
        for s in jsonObject['_systemList']:
            for p in s['_processorList']:
                tempProcessor = processor(name=p['_name'],description=p['_description'])
                tempProcessor.runs_rap_backbone=p['rap_backbone']
                tempProcessor.IP=p['_IP']
                # processor bindings
                for binding in p['_bindingList']:
                    # find the system and append
                    for sys in self.systems:
                        if sys.name == s["_name"]:
                            for comp in sys.processes:
                                if binding["_name"] == comp.name:
                                    tempProcessor.addProcessorBinding(process=comp)
                #processor properties
                for prop in p["_propertyList"]:
                    tempProp = ""
                    #TODO: extend the AADLIL TO SUPPORT PROCESSOR PROPERTIES

                #processor features
                for f in p['_featureList']:
                    _m = None
                    for m in self.messages:
                        if f['_message'] is not None:
                            if m.name == f['_message']['_name']:
                                _m = m
                    if f['_featureType'] == 'inport':
                        tempF = inport(name=f['_name'], type=f['_type'], message=_m)
                    elif f['_featureType'] == 'outport':
                        tempF = outport(name=f['_name'], type=f['_type'], message=_m)
                    elif f['_featureType'] == 'port':
                        tempF = outport(name=f['_name'], type=f['_type'],initialValue=f["_initialValue"],valueReference=f["_valueReference"], message=_m)
                    else:
                        tempF = ''
                    tempProcessor.addFeature(feature=tempF)
                for _s in self.systems:
                    if _s.name == s["_name"]:
                        _s.addProcessor(processor=tempProcessor)

    def __eq__(self, other):
        if not isinstance(other, system):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.__dict__ == other.__dict__



import re
import threading
import os
import io
import zipfile
from subprocess import Popen, CREATE_NEW_CONSOLE
import subprocess
import xml.etree.ElementTree as ET
import paho.mqtt.client as mqtt
import yaml
import redis


def getCustomCode(text,tag):
    pattern = r"#<!-- cc_"+tag+" START--!>(.*?)#<!-- cc_"+tag+" END--!>"
    matches = re.findall(pattern, text, re.DOTALL)
    return matches if matches else None

def replaceCustomCode(text,tag,replacement):
    pattern = r"#<!-- cc_"+tag+" START--!>(.*?)#<!-- cc_"+tag+" END--!>"
    start_tag = "#<!-- cc_"+tag+" START--!>"
    end_tag = "#<!-- cc_" + tag + " END--!>"
    return re.sub(pattern,start_tag+replacement[0]+end_tag,text,flags=re.DOTALL)


#----------------------------------------------------------------------------------------------------------------------
#------------------------------------------------CLI FUNCTIONS---------------------------------------------------------
#----------------------------------------------------------------------------------------------------------------------
def run_command(command):
    try:
        # result = subprocess.run(command[0][0], shell=True, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        process = Popen(command[0], shell=True, cwd=command[1])
        stdout, stderr = process.communicate()
        print(process.stdout)
    except subprocess.CalledProcessError as e:
        print(f"Command: {command}\nError: {e.stderr.decode('utf-8')}")


def execute_commands(commands):
    """
    Execute multiple command line functions concurrently.

    Parameters:
    commands (list of str): List of command line commands to execute.
    """
    threads = []
    for command in commands:
        thread = threading.Thread(target=run_command, args=(command,))
        thread.start()
        threads.append(thread)

    for thread in threads:
        thread.join()

def extractCommands(launchDescription):
    commands = []
    for component in launchDescription.components:
        commands.append([component.cmd, component.path])
    return commands


#----------------------------------------------------------------------------------------------------------------------
#--------------------------------------------LAUNCH FILE FUNCTIONS-----------------------------------------------------
#----------------------------------------------------------------------------------------------------------------------

# Example usage
xml_string = """
<launch>
    <node name="monitor" path="C:/Users/Bert/UAntwerpen/Documenten/00_UA/02_Projects/01_RoboSAPIENS/98_Sandbox/Redis/pythonProject/PyLauncher/monitor"/>
    <node name="analysis" path="C:/Users/Bert/UAntwerpen/Documenten/00_UA/02_Projects/01_RoboSAPIENS/98_Sandbox/Redis/pythonProject/PyLauncher/analysis"/>
</launch>
"""

class Component:
    def __init__(self, name, path,formalism):
        self.name = name
        self.path = path
        if formalism == 'python':
            self.cmd= ['python', name+'.py']
        if formalism == 'c++':
            self.cmd = [path+'/'+name+'.exe']

    def __repr__(self):
        return f"Swc(name='{self.name}', cmd='{self.cmd}')"

class Launch:
    def __init__(self, nodes):
        self.components = nodes

    def __repr__(self):
        return f"Launch(components={self.components})"
def parse_launch_xml(file,formalism="python"):
    with open(file, 'r') as f:
        data = f.read()
        root = ET.fromstring(data)
        components = []
        for node_elem in root.findall('node'):
            name = node_elem.get('name')
            path = node_elem.get('path')
            components.append(Component(name, path,formalism))
        return Launch(components)


#----------------------------------------------------------------------------------------------------------------------
#-------------------------------------------FILE HANDLING FUNCTIONS----------------------------------------------------
#----------------------------------------------------------------------------------------------------------------------
def decompress_folder(data, output_path):
    # Decompress the byte stream into the output folder
    with zipfile.ZipFile(io.BytesIO(data), 'r') as zip_file:
        zip_file.extractall(output_path)
    print(f"Folder decompressed to '{output_path}'.")

def compress_folder(folder_path):
    # Compress the folder into a byte stream
    zip_buffer = io.BytesIO()
    with zipfile.ZipFile(zip_buffer, 'w', zipfile.ZIP_DEFLATED) as zip_file:
        for root, _, files in os.walk(folder_path):
            for file in files:
                file_path = os.path.join(root, file)
                zip_file.write(file_path, os.path.relpath(file_path, folder_path))
    zip_buffer.seek(0)
    return zip_buffer.read()


#----------------------------------------------------------------------------------------------------------------------
#-------------------------------------------PYTHON ENVIRONMENT SETUP---------------------------------------------------
#----------------------------------------------------------------------------------------------------------------------
def get_activate_script_path(venv_name):
    """
    Returns the path to the activate script based on the operating system.
    """
    return os.path.join(venv_name, "Scripts", "activate.bat") if os.name == "nt" else os.path.join(venv_name, "bin", "activate")


def get_pip_path(venv_name):
    """
    Returns the path to the pip executable based on the operating system.
    """
    return os.path.join(venv_name, "Scripts", "pip.exe") if os.name == "nt" else os.path.join(venv_name, "bin", "pip")


def create_virtual_environment(venv_name="venv"):
    """
    Creates a Python 3.10 virtual environment.

    :param venv_name: The name of the virtual environment directory. Defaults to "venv".
    """
    if os.path.exists(venv_name):
        print(f"Virtual environment '{venv_name}' already exists. Skipping creation.")
        return

    try:
        # Check if Python 3.10 is installed
        python_version_check = subprocess.run(["python", "--version"], capture_output=True, text=True)

        # Install virtualenv package if not installed
        subprocess.run(["python", "-m", "pip", "install", "virtualenv"], check=True)

        # Create virtual environment using virtualenv
        subprocess.run(["python", "-m", "virtualenv", venv_name, "--python=python3.10"], check=True)
        print(f"Virtual environment '{venv_name}' created successfully.")
    except subprocess.CalledProcessError as e:
        print(f"Error occurred while creating virtual environment: {e}")


def activate_virtual_environment(venv_name="venv"):
    """
    Activates the virtual environment.

    :param venv_name: The name of the virtual environment directory. Defaults to "venv".
    """
    activate_script = get_activate_script_path(venv_name)
    if not os.path.exists(activate_script):
        print(f"Activate script not found in the virtual environment '{venv_name}'. Make sure the virtual environment is created.")
        return

    if os.name == "nt":
        subprocess.run(activate_script, shell=True)
    else:
        subprocess.run(["source", activate_script], shell=True, executable="/bin/bash")


def deactivate_virtual_environment():
    """
    Deactivates the virtual environment.
    """
    if os.name == "nt":
        subprocess.run("deactivate", shell=True)
    else:
        subprocess.run("deactivate", shell=True, executable="/bin/bash")


def install_requirements(venv_name="venv", requirements_file="requirements.txt"):
    """
    Installs packages listed in a requirements file into the virtual or native environment .

    :param venv_name: The name of the virtual environment directory. Defaults to "venv".
    :param requirements_file: The path to the requirements.txt file. Defaults to "requirements.txt".
    """
    if venv_name is not None:
        pip_path = get_pip_path(venv_name)
        if not os.path.exists(pip_path):
            print(f"Pip not found in the virtual environment '{venv_name}'. Make sure the virtual environment is created.")
            return
    else:
        pip_path = "pip"
        print(f"Using pip of native python environment.")

    if not os.path.isfile(requirements_file):
        print(f"Requirements file '{requirements_file}' not found.")
        return

    try:
        # Install requirements
        subprocess.run([pip_path, "install", "-r", requirements_file], check=True)
        print(f"Packages from '{requirements_file}' installed successfully.")
    except subprocess.CalledProcessError as e:
        print(f"Error occurred while installing requirements: {e}")

def get_python_version():
    """
    Checks and returns the current Python version installed on the system.

    :return: A string representing the Python version, or None if an error occurs.
    :rtype: str or None
    """
    try:
        # Run the command to get the Python version
        result = subprocess.run(["python", "--version"], capture_output=True, text=True, check=True)

        # Output is usually in the form of "Python X.Y.Z\n"
        version = result.stdout.strip()

        return version
    except subprocess.CalledProcessError as e:
        print(f"Error occurred while checking Python version: {e}")
        return None

#----------------------------------------------------------------------------------------------------------------------
#--------------------------------------------DOCKER SETUP FUNCTIONS----------------------------------------------------
#----------------------------------------------------------------------------------------------------------------------
def build_docker_image(module_path, image_name):
    """
    Build a Docker image for a Python module at a given path.

    :param module_path: Path to the module (should contain Dockerfile)
    :param image_name: Name of the Docker image to be created
    :return: None
    """
    if not os.path.isdir(module_path):
        raise FileNotFoundError(f"The specified module path '{module_path}' does not exist or is not a directory.")

    dockerfile_path = os.path.join(module_path, 'Dockerfile')
    if not os.path.isfile(dockerfile_path):
        raise FileNotFoundError(f"No Dockerfile found in the specified module path '{module_path}'.")

    try:
        # Build the Docker image using the specified Dockerfile
        command = ["docker", "build", "-t", image_name, module_path]
        subprocess.run(command, check=True)
        print(f"Successfully built Docker image '{image_name}' from '{module_path}'.")
    except subprocess.CalledProcessError as e:
        print(f"Failed to build Docker image. Error: {e}")


def run_docker_container(image_name, container_name=None, ports=None):
    """
    Run a Docker container from an existing image.

    :param image_name: Name of the Docker image to run
    :param container_name: Optional name for the container
    :param ports: Optional dictionary mapping container ports to host ports (e.g., {"8080": "8080"})
    :return: None
    """
    try:
        # Prepare the docker run command
        command = ["docker", "run", "-d"]

        if container_name:
            command.extend(["--name", container_name])

        if ports:
            for host_port, container_port in ports.items():
                command.extend(["-p", f"{host_port}:{container_port}"])

        command.append(image_name)

        # Run the Docker container
        subprocess.run(command, check=True)
        print(f"Successfully started container from image '{image_name}'.")
    except subprocess.CalledProcessError as e:
        print(f"Failed to run Docker container. Error: {e}")


def get_docker_version():
    """
    Check if Docker is installed on the system

    :return: String with docker version if Docker is installed, False otherwise
    """
    try:
        result = subprocess.run(["docker", "--version"], capture_output=True, text=True, check=True)
        version = result.stdout.strip()
        return version
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False


#----------------------------------------------------------------------------------------------------------------------
#----------------------------------------------ENVIRONMENT CHECKS------------------------------------------------------
#----------------------------------------------------------------------------------------------------------------------
def check_redis(host="localhost", port=6379, db=0, timeout=30,config=None):
    """
    Check if Redis is running and reachable

    :return: True if Redis is running and reachable, False otherwise
        """

    try:
        if config is None:
            print(
                'WARNING: configuration file not provided, checking Redis with default values (broker="localhost", port=6379, db=0)')
        else:
            with open(config, 'r') as file:
                configuration = yaml.safe_load(file)
                host = configuration['redis_host']
                port = configuration['redis_port']

        # Establish a connection with the specified timeout
        client = redis.Redis(host=host, port=port, db=db, socket_timeout=timeout)

        # Ping the server to test connectivity
        if client.ping():
            return True  # Redis is reachable
    except redis.exceptions.ConnectionError:
        pass  # Connection failed

    return False  # Redis is not reachable

def check_mqtt(broker="localhost", port=1883, timeout=30,config=None):
    """
    Check if MQTT is running and reachable

    :return: True if MQTT broker is running and reachable, False otherwise
        """

    def on_connect(client, userdata, flags, rc):
        # If rc (return code) is 0, connection was successful
        client.reachable = (rc == 0)
        client.disconnect()

    client = mqtt.Client()
    client.reachable = False  # Initial assumption: not reachable
    client.on_connect = on_connect

    #resolve the config file for checking the MQTT config
    if config is None:
        print('WARNING: configuration file not provided, checking MQTT with default values (broker="localhost", port=1833)')
    else:
        with open(config, 'r') as file:
            configuration=yaml.safe_load(file)
            broker = configuration['mqtt_broker']
            port = configuration['mqtt_port']

    # Attempt to connect with specified timeout
    try:
        client.connect(broker, port, timeout)
        client.loop_start()  # Start the loop to process callbacks
        client.loop(timeout)  # Wait for connection
        client.loop_stop()  # Stop the loop
    except Exception as e:
        print(f"Could not connect to MQTT broker: {e}")

    return client.reachable



# Copyright (c) 2023-present Bert Van Acker (UA) <bert.vanacker@uantwerpen.be>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from rpio.transformations.transformations import *
from rpio.parsers.parsers import *
from rpio.metamodels.aadl2_IL import *




@click.group()
@click.pass_context
def transformationCmds():
    pass

@transformationCmds.command()
@click.option('--verbose','-v', is_flag=True,default=False,help='Enable debug information.')
@click.option('--roboarch2aadl', is_flag=True,default=False,help='Perform the roboArch2AADL transformation.')
@click.option('--aadl2aadlil', is_flag=True,default=False,help='Perform the AADL2AADLIL transformation.')
@click.option('--aadlil2code', is_flag=True,default=False,help='Perform the AADLIL2CODE transformation.')
def transformation(verbose,roboarch2aadl,aadl2aadlil,aadlil2code):
    """Performing one of the model-to-model or model-to-code transformations."""

    if roboarch2aadl:
        if verbose:print("WARNING: RoboArch2AADL transformation is not implemented yet.")

    if aadl2aadlil:
        if verbose: print("WARNING: AADL2AADLIL transformation is under development.")

        # 1. Setup the AADL parser
        try:
            parser = AADL_parser(logicalArchitecture='Design/logicalArchitecture.aadl',physicalArchitecture='Design/physicalArchitecture.aadl', messages='Design/messages.aadl')
        except:
            parser = None
            if verbose:print("ERROR: AADL parser could not be instantiated, check if AADL models are available (Design/*.aadl).")


        # 2. parse the aadl models and store in AADLIL
        try:
            s = parser.aadl2aadlIl()
        except:
            s = None
            if verbose:print("ERROR: AADL parsing failed, check if AADL models are available (Design/*.aadl).")

        # 3. dump to aadlil json
        try:
            s.object2json(fileName='Design/design.json')
        except:
            if verbose:print("ERROR: AADLIL model could not be generated.")


    if aadlil2code:
        if verbose: print("WARNING: AADL2AADLIL transformation is under development.")

        # 1. LOAD THE AADL INTERMEDUATE LANGUAGE
        try:
            design = system(name="adaptiveSystem", description="Design generated from the AADLIL file",JSONDescriptor='Design/design.json')
        except:
            design = None
            if verbose:print("ERROR: The AADIL file could not be loaded, please check if it exists (Design/design.json).")

        # 2. GENERATE CUSTOM MESSAGES FROM AADL INTERMEDIATE LANGUAGE
        try:
            message2code_py(system=design, path="Realization/ManagingSystem/Messages")
            message2code_py(system=design, path="Realization/ManagedSystem/Messages")
        except:
            if verbose:print("ERROR: Messages could not be generated, no design loaded.")

        # 3. GENERATE SWC CODE FROM AADL INTERMEDIATE LANGUAGE
        try:
            swc2code_py(system=design, path="Realization/ManagingSystem/Nodes")
        except:
            if verbose:print("ERROR: Code could not be generated, no design loaded.")

        # 3. GENERATE PLATFORM LAUNCH FILES
        try:
            swc2launch(system=design.systems[0], path="Realization/ManagingSystem/Platform")
            swc2launch(system=design.systems[1], path="Realization/ManagedSystem/Platform")
        except:
            if verbose: print("ERROR: Platform-specific launch file could not be generated, no design loaded.")

        # 3. GENERATE PLATFORM MAIN FILES
        try:
            current_folder_path, current_folder_name = os.path.split(os.getcwd())
            swc2main(system=design.systems[0],package=current_folder_name,prefix=None,path="Resources")
        except:
            if verbose: print("ERROR: Platform(s) main file could not be generated.")

        # 4. GENERATE PLATFORM DOCKER COMPOSE
        try:
            swc2dockerCompose(system=design.systems[0], path="Realization/ManagingSystem/Platform")
            add_backbone_config(system=design, path="Resources")
        except:
            if verbose: print("ERROR: Platform(s) docker compose file could not be generated.")

        # 8. update the roboSapiensIO.ini file based on the generation
        update_robosapiensIO_ini(system=design, path=None)


#**********************************************************************************
# * Copyright (C) 2024-present Bert Van Acker (B.MKR) <bert.vanacker@uantwerpen.be>
# *
# * This file is part of the roboarch R&D project.
# *
# * RAP R&D concepts can not be copied and/or distributed without the express
# * permission of Bert Van Acker
# **********************************************************************************
import os
import re
import sys
from rpio.metamodels.aadl2_IL import *
from textx import metamodel_from_file

class AADL_parser:
    """This is a parser that parses muliple AADL files and puts them into the AADL intermediate language.

        :param [logicalArchitecture]: [Path to the logical architecture AADL model], defaults to [None]
        :type [logicalArchitecture]: [string](, optional)

        :param [logicalArchitecture]: [Path to the logical architecture AADL model], defaults to [None]
        :type [logicalArchitecture]: [string](, optional)

        :param [messages]: [Path to the messages AADL model], defaults to [None]
        :type [messages]: [string](, optional)

        """
    def __init__(self, logicalArchitecture,physicalArchitecture,messages):
        """Constructor method
        """
        self.adaptiveSystem = None
        self.managingSystem = None
        self.managedSystem = None

        self.messages = []
        self.processes = []

        # Read the logical architecture
        try:
            with open(logicalArchitecture, 'r') as file:
                self.logicalArchitectureAADL = file.read()
        except FileNotFoundError:
            self.logicalArchitectureAADL = None
            print(f"Error: File '{logicalArchitecture}' not found.")
            sys.exit(1)

        # Read the physical architecture
        try:
            with open(physicalArchitecture, 'r') as file:
                self.physicalArchitectureAADL = file.read()
        except FileNotFoundError:
            self.physicalArchitectureAADL = None
            print(f"Error: File '{physicalArchitecture}' not found.")
            #sys.exit(1)        #TODO: enable if implemented

        # Read the messages
        try:
            with open(messages, 'r') as file:
                self.messagesAADL = file.read()
        except FileNotFoundError:
            self.messagesAADL = None
            print(f"Error: File '{messages}' not found.")
            sys.exit(1)




    def aadl2aadlIl(self):
        """Function to parse the AADL models and put them into the AADL intermediate language

        :return: [adaptiveSystem]
        :rtype: [Object]
        """

        # 0. Setup adaptive system
        self.adaptiveSystem = system(name="adaptiveSystem", description="Generated from AADL models")

        #1. Setup managing and managed system
        self.managedSystem = system(name="managedSystem", description="managed system part")
        self.managingSystem = system(name="managingSystem", description="managing system part")

        # 1. Parse the messages
        self._generateMessages()
        self.adaptiveSystem.messages = self.messages

        # 2. Parse the processes
        self._generateProcesses()

        # 3. Populate adaptive system
        self.adaptiveSystem.addSystem(self.managingSystem)
        self.adaptiveSystem.addSystem(self.managedSystem)



        return self.adaptiveSystem


    def _generateProcesses(self):
        """Function to parse the AADL processes."""

        # Find all matches with "process" followed by any word
        pattern = r'process\s+([\w:]+)'
        matches = re.findall(pattern, self.logicalArchitectureAADL, re.DOTALL)
        # Filter out matches containing "implementation"
        filtered_matches = [match for match in matches if "implementation" not in match]
        # Generate process components for each of the matches
        for match in filtered_matches:
            p = process(name=match, description=match+" component")

            # add features to the process
            pattern = r"process " + match + "(.*?)end " + match + ";"
            matches = re.findall(pattern, self.logicalArchitectureAADL, re.DOTALL)
            feature_pattern = r'(\w+):\s+(in|out)\s+(event|data|event data)\s+port\s+([\w:]+);'
            matches = re.findall(feature_pattern, matches[0], re.DOTALL)
            for feature in matches:
                # Lambda function to find an object by name
                find_by_name = lambda name: next((item for item in self.messages if item.name == name), None)

                if feature[1] == "in":
                    f = inport(name=feature[0], type=feature[2], message=find_by_name(feature[3].split("::")[1]))
                    p.addFeature(f)
                elif feature[1] == "out":
                    f = outport(name=feature[0], type=feature[2], message=find_by_name(feature[3].split("::")[1]))
                    p.addFeature(f)

            # add threads to the process
            try:
                pattern = r"process implementation " + match + ".impl(.*?)end " + match + ".impl;"
                matches = re.findall(pattern, self.logicalArchitectureAADL, re.DOTALL)
                thread_pattern = r'(\w+): thread\s+([\w:]+);'
                threadComponents = re.findall(thread_pattern, matches[0], re.DOTALL)
                featureList=[]
                #fetch thread content
                for t in threadComponents:
                    pattern = r"thread " + t[0] + "(.*?)end " + t[0] + ";"
                    matches = re.findall(pattern, self.logicalArchitectureAADL, re.DOTALL)
                    matches = re.findall(feature_pattern, matches[0], re.DOTALL)
                    for feature in matches:
                        # Lambda function to find an object by name
                        find_by_name = lambda name: next((item for item in self.messages if item.name == name), None)

                        if feature[1] == "in":
                            f = inport(name=feature[0], type=feature[2],message=find_by_name(feature[3].split("::")[1]))
                            featureList.append(f)
                        elif feature[1] == "out":
                            f = outport(name=feature[0], type=feature[2],message=find_by_name(feature[3].split("::")[1]))
                            featureList.append(f)

                th = thread(name=t[0], featureList=featureList)
                p.addThread(th)

            except:
                print("Process "+ match + " does not have an implementation")



            self.processes.append(p)
            # add to the managing system or managed system
            if 'monitor' in p.name or 'analysis' in p.name or 'plan' in p.name or 'legitimate' in p.name or 'execute' in p.name:        #TODO: check if this is ok for the user
                self.managingSystem.addProcess(p)
            elif 'knowledge' not in p.name:
                self.managedSystem.addProcess(p)
        return self.processes

    def _generateMessages(self):
        """Function to parse the AADL messages."""

        # Gind all matches with the "data" pattern
        pattern = r"data (\w+)(.*?)end \1;"
        matches = re.findall(pattern, self.messagesAADL, re.DOTALL)
        for match in matches:
            name = match[0]

            # --- SIMPLE MESSAGES, EXTRACT FEATURES
            feature_pattern = r"(\w+): provides data access ([\w:]+);"
            features = re.findall(feature_pattern, match[1])
            featureList = []
            for feature in features:
                if '::' in feature[1]:
                    f = data(name=feature[0], dataType=feature[1].split("::")[1])
                else:
                    f = data(name=feature[0], dataType=feature[1])
                featureList.append(f)
            m = message(name=name, featureList=featureList)
            self.messages.append(m)
        return self.messages


class robochart_parser:
    """This is a parser that parses muliple robochart files, as input for the AADL code generator.

        :param [MAPLEK]: [Path to the MAPLE-K roboChart model], defaults to [None]
        :type [MAPLEK]: [string](, optional)

        :param [Monitor]: [Path to the Monitor roboChart model], defaults to [None]
        :type [Monitor]: [string](, optional)

        :param [Analysis]: [Path to the Analysis roboChart model], defaults to [None]
        :type [Analysis]: [string](, optional)

        :param [Plan]: [Path to the Plan roboChart model], defaults to [None]
        :type [Plan]: [string](, optional)

        :param [Legitimate]: [Path to the Legitimate roboChart model], defaults to [None]
        :type [Legitimate]: [string](, optional)

        :param [Execute]: [Path to the Execute roboChart model], defaults to [None]
        :type [Execute]: [string](, optional)

        :param [Knowledge]: [Path to the Knowledge roboChart model], defaults to [None]
        :type [Knowledge]: [string](, optional)

        """
    def __init__(self, MAPLEK,Monitor,Analysis,Plan,Legitimate,Execute,Knowledge):
        """Constructor method
        """
        currentDir = os.path.dirname(os.path.abspath(__file__))
        metamodelPath = os.path.join(currentDir, 'robochart/robochart.tx')
        self.robochart_meta = metamodel_from_file(metamodelPath)

        self.maplek_model = None
        self.monitor_model = None
        self.analysis_model = None
        self.plan_model = None
        self.legitimate_model = None
        self.execute_model = None
        self.knowledge_model = None

        # Read the MAPLE-K robochart model
        try:
            self.maplek_model = self.robochart_meta.model_from_file(MAPLEK)
        except FileNotFoundError:
            self.maplek_model = None
            print(f"Error: File '{MAPLEK}' not found.")
            sys.exit(1)

        # Read the Monitor robochart model
        try:
            self.monitor_model = self.robochart_meta.model_from_file(Monitor)
        except FileNotFoundError:
            self.monitor_model = None
            print(f"Error: File '{Monitor}' not found.")
            sys.exit(1)

        # Read the Analysis robochart model
        try:
            self.analysis_model = self.robochart_meta.model_from_file(Analysis)
        except FileNotFoundError:
            self.analysis_model = None
            print(f"Error: File '{Analysis}' not found.")
            sys.exit(1)

        # Read the Plan robochart model
        try:
            self.plan_model = self.robochart_meta.model_from_file(Plan)
        except FileNotFoundError:
            self.plan_model = None
            print(f"Error: File '{Plan}' not found.")
            sys.exit(1)

        # Read the Legitimate robochart model
        try:
            self.legitimate_model = self.robochart_meta.model_from_file(Legitimate)
        except FileNotFoundError:
            self.legitimate_model = None
            print(f"Error: File '{Legitimate}' not found.")
            sys.exit(1)

        # Read the Execute robochart model
        try:
            self.execute_model = self.robochart_meta.model_from_file(Execute)
        except FileNotFoundError:
            self.execute_model = None
            print(f"Error: File '{Execute}' not found.")
            sys.exit(1)

        # Read the Knowledge robochart model
        try:
            self.knowledge_model = self.robochart_meta.model_from_file(Knowledge)
        except FileNotFoundError:
            self.knowledge_model = None
            print(f"Error: File '{Knowledge}' not found.")
            sys.exit(1)